@@ -2,15 +2,15 @@
 #
 # miniirc - A small-ish IRC framework.
 #
-# © 2018-2020 by luk3yx and other contributors of miniirc.
+# © 2018-2022 by luk3yx and other contributors of miniirc.
 #
 
-import atexit, errno, threading, time, socket, ssl, sys
+import atexit, errno, threading, time, select, socket, ssl, sys, warnings
 
 # The version string and tuple
-ver = __version_info__ = (1,7,0)
-version = 'miniirc IRC framework v1.7.0'
-__version__ = '1.7.0'
+ver = __version_info__ = (1,8,3)
+version = 'miniirc IRC framework v1.8.3'
+__version__ = '1.8.3'
 
 # __all__ and _default_caps
 __all__ = ['CmdHandler', 'Handler', 'IRC']
@@ -27,8 +27,15 @@ except ImportError:
 
 # Create global handlers
 _global_handlers = {}
+_colon_warning = False
 
 def _add_handler(handlers, events, ircv3, cmd_arg, colon):
+    if (colon and _colon_warning and
+            not all(str(e).upper().startswith('IRCV3 ') for e in events)):
+        warnings.warn('Using colon=True or not specifying the colon '
+                      'keyword argument to miniirc.Handler is deprecated.',
+                      DeprecationWarning, stacklevel=3)
+
     if not events:
         if not cmd_arg:
             raise TypeError('Handler() called without arguments.')
@@ -177,7 +184,7 @@ class IRC:
     debug_file = sys.stdout
     sendq = None
     msglen = 512
-    _main_lock = None
+    _main_thread = None
     _sasl = False
     _unhandled_caps = None
 
@@ -276,10 +283,24 @@ class IRC:
         if tags:
             msg = _dict_to_tags(tags) + msg
 
+        msg += b'\r\n'
         self._send_lock.acquire()
         try: # Apparently try/finally is faster than "with".
-            self.sock.sendall(msg + b'\r\n')
-        except (AttributeError, BrokenPipeError):
+            while True:
+                try:
+                    self.sock.sendall(msg)
+                    break
+                except (BlockingIOError, ssl.SSLWantReadError):
+                    # Wait for the socket to become ready again
+                    readable, _, _ = select.select(
+                        (self.sock,), (), (self.sock,),
+                        self.ping_timeout or self.ping_interval
+                    )
+                except ssl.SSLWantWriteError:
+                    select.select((), (self.sock,), (self.sock,),
+                                  self.ping_timeout or self.ping_interval)
+        except (AttributeError, BrokenPipeError, socket.timeout):
+            # TODO: Consider not silently ignoring timeouts
             if force:
                 raise
         finally:
@@ -325,21 +346,21 @@ class IRC:
             self._send_lock.release()
 
         self.debug('Connecting to', self.ip, 'port', self.port)
-        addrinfo = socket.getaddrinfo(self.ip, self.port, 0,
-            socket.SOCK_STREAM)[0]
-        self.sock = socket.socket(*addrinfo[:2])
+        self.sock = socket.create_connection(
+            (self.ip, self.port),
+            timeout=self.ping_timeout or self.ping_interval,
+        )
         if self.ssl:
             self.debug('SSL handshake')
+            ctx = ssl.create_default_context(cafile=get_ca_certs())
             if self.verify_ssl:
-                self.sock = ssl.wrap_socket(self.sock,
-                    cert_reqs=ssl.CERT_REQUIRED, ca_certs=get_ca_certs(),
-                    do_handshake_on_connect=True)
+                assert ctx.check_hostname
             else:
-                self.sock = ssl.wrap_socket(self.sock,
-                    do_handshake_on_connect=True)
-        self.sock.connect(addrinfo[4])
-        if self.ssl and self.verify_ssl:
-            ssl.match_hostname(self.sock.getpeercert(), self.ip)
+                warnings.warn('Disabling verify_ssl is usually a bad idea.')
+                ctx.check_hostname = False
+                ctx.verify_mode = ssl.CERT_NONE
+            self.sock = ctx.wrap_socket(self.sock, server_hostname=self.ip)
+
         # Begin IRCv3 CAP negotiation.
         self._unhandled_caps = None
         self.quote('CAP LS 302', force=True)
@@ -349,7 +370,14 @@ class IRC:
         atexit.register(self.disconnect)
         self.debug('Starting main loop...')
         self._sasl = self._pinged = False
-        self.main()
+        self._start_main_loop()
+
+    def _start_main_loop(self):
+        # Start the thread before updating _main_thread so that
+        # wait_until_disconnected() works correctly.
+        thread = threading.Thread(target=self._main)
+        thread.start()
+        self._main_thread = thread
 
     # Disconnect from IRC.
     def disconnect(self, msg=None, *, auto_reconnect=False):
@@ -431,34 +459,54 @@ class IRC:
 
     # The main loop
     def _main(self):
+        # Make the socket non-blocking.
+        self.sock.setblocking(False)
+
         self.debug('Main loop running!')
         buffer = b''
         while True:
             try:
                 assert len(buffer) < 65535, 'Very long line detected!'
                 try:
-                    raw = self.sock.recv(8192).replace(b'\r', b'\n')
+                    # Acquire the send lock when receiving data because I don't
+                    # think you're supposed to call SSL functions from multiple
+                    # threads at once
+                    self._send_lock.acquire()
+                    try:
+                        raw = self.sock.recv(8192).replace(b'\r', b'\n')
+                    finally:
+                        self._send_lock.release()
+
                     if not raw:
                         raise ConnectionAbortedError
                     buffer += raw
-                except socket.timeout:
-                    if self._pinged:
-                        raise
-                    else:
+                except (BlockingIOError, ssl.SSLWantReadError):
+                    # Wait for the socket to become ready again
+                    readable, _, _ = select.select(
+                        (self.sock,), (), (self.sock,),
+
+                        # self.ping_interval should be used when
+                        # self.ping_timeout is None
+                        (self._pinged and self.ping_timeout or
+                         self.ping_interval)
+                    )
+
+                    # Handle ping timeouts
+                    if not readable:
+                        if self._pinged:
+                            raise TimeoutError
                         self._pinged = True
-                        if self.ping_timeout:
-                            self.sock.settimeout(self.ping_timeout)
                         self.quote('PING', ':miniirc-ping', force=True)
-                except socket.error as e:
-                    if e.errno != errno.EWOULDBLOCK:
-                        raise
+                except ssl.SSLWantWriteError:
+                    select.select((), (self.sock,), (self.sock,),
+                                  self.ping_timeout or self.ping_interval)
+
             except (OSError, socket.error) as e:
                 self.debug('Lost connection!', repr(e))
                 self.disconnect(auto_reconnect=True)
                 while self.persist:
                     time.sleep(5)
                     self.debug('Reconnecting...')
-                    self._main_lock = None
                     try:
                         self.connect()
                     except (OSError, socket.error):
@@ -485,15 +533,19 @@ class IRC:
                         self.debug('Ignored message:', line)
             del raw
 
-    # Thread the main loop
+    def wait_until_disconnected(self, *, _timeout=None):
+        # The main thread may be replaced on reconnects
+        while self._main_thread and self._main_thread.is_alive():
+            self._main_thread.join(_timeout)
+
     def main(self):
-        if self._main_lock and self._main_lock.is_alive():
-            self.debug('Main loop already running!')
-            return self._main_lock
-        self.debug('Creating new thread...')
-        self._main_lock = threading.Thread(target=self._main)
-        self._main_lock.start()
-        return self._main_lock
+        warnings.warn('The miniirc.IRC.main() function is deprecated and '
+                      'should not be used.', DeprecationWarning, 2)
+        # The main thread may be started after the if check and before
+        # _start_main_loop. This function is deprecated so fixing this probably
+        # isn't worthwhile.
+        if not self._main_thread or not self._main_thread.is_alive():
+            self._start_main_loop()
 
 # Handle some IRC messages by default.
 @Handler('001')
@@ -501,8 +553,6 @@ def _handler(irc, hostmask, args):
     irc.connected = True
     irc.isupport.clear()
     irc._unhandled_caps = None
-    if irc.ping_interval:
-        irc.sock.settimeout(irc.ping_interval)
     irc.debug('Connected!')
     if irc.connect_modes:
         irc.quote('MODE', irc.nick, irc.connect_modes)
@@ -527,8 +577,6 @@ def _handler(irc, hostmask, args):
 def _handler(irc, hostmask, args):
     if args and args[-1] == 'miniirc-ping' and irc.ping_interval:
         irc._pinged = False
-        if irc.ping_timeout:
-            irc.sock.settimeout(irc.ping_interval)
 
 @Handler('432', '433')
 def _handler(irc, hostmask, args):
@@ -617,8 +665,9 @@ def _handler(irc, hostmask, args):
 
 @Handler('904', '905')
 def _handler(irc, hostmask, args):
-    irc._sasl = False
-    irc.quote('AUTHENTICATE *', force=True)
+    if irc._sasl:
+        irc._sasl = False
+        irc.quote('AUTHENTICATE *', force=True)
 
 @Handler('902', '903', '904', '905')
 def _handler(irc, hostmask, args):
@@ -633,10 +682,13 @@ def _handler(irc, hostmask, args):
         except (IndexError, ValueError):
             return
 
+        # Stop irc.wait_until_disconnected() from returning early
+        irc._main_thread = threading.current_thread()
+
         persist = irc.persist
         irc.disconnect()
-        irc.debug('NOTICE: An IRCv3 STS has been detected, the port will',
-            'be changed to', port, 'and TLS/SSL will be enabled.')
+        irc.debug('STS detected, enabling TLS/SSL and changing the port to ',
+                  port)
         irc.port = port
         irc.ssl = True
         time.sleep(1)
@@ -656,11 +708,9 @@ def _handler(irc, hostmask, args):
     irc.finish_negotiation(args[0])
 
 # Handle ISUPPORT messages
-@Handler('005', colon=True)
+@Handler('005')
 def _handler(irc, hostmask, args):
-    del args[0]
-    if args[-1].startswith(':'):
-        del args[-1]
+    del args[0], args[-1]
     isupport = _tags_to_dict(args, None)
 
     # Try and auto-detect integers
@@ -678,4 +728,5 @@ def _handler(irc, hostmask, args):
 
     irc.isupport.update(isupport)
 
+_colon_warning = True
 del _handler
